#!/usr/bin/python
import os, shutil, stat, time, subprocess, sys
from collections import OrderedDict

# a dict with some useful additional getters which can convert types and handle one-element lists like their single member
class ConfigDict(dict):
	def getstr(self, name, default = None):
		if not name in self: return default
		val = self[name]
		if isinstance(val, list):
			if len(val) != 1: raise Exception('%s is a list, but it should not' % name)
			return val[0]
		else:
			return val
	
	def getint(self, name, default = None):
		return int(self.getstr(name, default))
	
	def getbool(self, name, default = None):
		val = self.getstr(name, default)
		if isinstance(val, bool): return val # already a bool
		return val.lower() in ('true', 'yes', 'on', '1')

# create a safe-to-call shell command from the array
def safeCall(*args):
	res = ""
	for arg in args:
		assert arg.find("'") < 0 # ' is not supported
		if len(res): res += " "
		res += "'"+arg+"'"
	return res

# Load a section-less config file: maps parameter names to space-separated lists of strings (with shell quotation)
# Lines starting with spaces are continuation lines
def loadConfigFile(file):
	import shlex
	# read config file
	linenr = 0
	with open(file) as file:
		result = ConfigDict()
		curKey = None
		for line in file:
			linenr += 1
			isCont = len(line) and line[0].isspace() # remember if we were a continuation line
			if isCont and curKey is None:
				raise Exception("Invalid config, line %d: Starting with continuation line" % linenr)
			line = line.strip()
			if not len(line) or line.startswith("#"): continue # skip empty and comment lines
			try:
				if isCont:
					# continuation line
					result[curKey] += shlex.split(line)
				else:
					# option line
					pos = line.index("=") # will raise exception when substring is not found
					curKey = line[:pos].strip()
					value = line[pos+1:]
					result[curKey] = shlex.split(value)
			except Exception:
				raise Exception("Invalid config, line %d: Error parsing line (quoting issue?)" % linenr)
	# add some convencience get functions
	return result

# representation of a build system
class BuildSystem:
	def __init__(self, ruleMaker, buildDepends = [], binaryDepends = []):
		self.ruleMaker = ruleMaker
		self.buildDepends = buildDepends
		self.binaryDepends = binaryDepends

# abstract representation of rules file
class RulesFile:
	def __init__(self):
		self.env = {}
		self.dh = []
		self.dhWith = set()
		self.rules = OrderedDict()
	
	def env2str(self):
		res = ""
		for name in self.env: # we rely on the name being sane (i.e., no special characters)
			val = self.env[name]
			assert val.find("'") < 0 # ' is not supported
			if len(res): res += " "
			res += name+"='"+val+"'"
		return res

	def write(self, f):
		print >>f, "#!/usr/bin/make -f"
		print >>f, ""
		print >>f, ".PHONY: build" # there may be a directory called "build"
		print >>f, ""
		print >>f, "build %:" # need to mention "build" here again explicitly so PHONY takes effect
		# write proper dh call
		dh = self.dh
		if self.dhWith:
			dh.append('--with='+','.join(self.dhWith))
		print >>f, '\t'+self.env2str()+' dh $@ '+safeCall(*dh)
		# write remaining rules
		for rule in self.rules:
			print >>f, ""
			print >>f, "override_dh_"+rule+":"
			for line in self.rules[rule]:
				print >>f, "\t"+line

# rule-makers
def cmakeRules(config):
	buildDir = config.getstr('buildDir', 'build')
	srcDir = os.getcwd()
	r = RulesFile()
	r.dh += ["--buildsystem=cmake", "--builddirectory="+buildDir] # dh parameters
	r.rules['auto_configure'] = [
		safeCall("mkdir", "-p", buildDir),
		safeCall("cd", buildDir) + " && " +
		  safeCall("cmake", srcDir, "-DCMAKE_INSTALL_PREFIX=/usr", *config.get('cmakeParameters', []))
	]
	r.rules['auto_clean'] = [safeCall('rm', '-f', os.path.join(buildDir, 'CMakeCache.txt'))] # clean old cmake cache
	return r

def automakeRules(config):
	# "build" is what we are building *on*, and "host" is what we are building *for* (and GNU is weird...)
	# also, we don't really support cross-building... ;-) (to do, we'd have to write shell code that checks whether BUILD_GNU_TYPE
	#  and HOST_GNU_TYPE are equal, and if they are not, add a --host parameter)
	r = RulesFile()
	r.dh += ["--buildsystem=autoconf"]
	r.rules['auto_configure'] = [
		'BUILD_TYPE=$$(dpkg-architecture -qDEB_BUILD_GNU_TYPE) && ' + # doing the expansion beforehand ensures that we cancel if it fails
		'MULTIARCH=$$(dpkg-architecture -qDEB_HOST_MULTIARCH) && '+
			'./configure --build=$$BUILD_TYPE '+ # do the escape manually here so we can use the variables (there's no user-controlled string in here anyway)
			'--prefix=/usr --includedir=/usr/include --mandir=/usr/share/man --infodir=/usr/share/info '+
			'--libdir=/usr/lib/$$MULTIARCH --libexecdir=/usr/lib/$$MULTIARCH '+
			'--sysconfdir=/etc --localstatedir=/var '+
			safeCall(*config.get('automakeParameters', []))
	]
	r.rules['auto_clean'] = ['rm -f config.status'] # do not re-use old configuration (no need for escaping here, obviously)
	return r

def pythonRules(config):
	r = RulesFile()
	r.dh += ["--buildsystem=python_distutils"]
	r.dhWith.add('python2')
	r.rules['auto_clean'] = [ # clean properly
		'dh_auto_clean',
		'rm -rf build'
	]
	return r

# build systems
buildSystems = {
	'cmake': BuildSystem(cmakeRules, ["cmake"]),
	'automake': BuildSystem(automakeRules),
	'python': BuildSystem(pythonRules, ["python-setuptools"], ["${python:Depends}"]),
}

# utility functions
def commandInBuildEnv(config, command):
	schroot = config.getstr('schroot')
	if schroot is not None: command = ['schroot', '-c', schroot, '--'] + command
	return command

def getArchitecture(config):
	cmd = commandInBuildEnv(config, ['dpkg-architecture', '-qDEB_HOST_ARCH'])
	p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
	res = p.communicate()[0] # get only stdout
	if p.returncode != 0: raise Exception("Querying dpkg for the architecture failed")
	return res[0:len(res)-1] # chop of the \n at the end

def writeDependency(f, name, list):
	if len(list):
		print >>f, name+": "+', '.join(list)

# actual work functions
def createDebianFiles(config):
	if not isinstance(config, ConfigDict):
		config = ConfigDict(config)
	sourceName = config.getstr('sourceName')
	binaryName = config.getstr('binaryName', sourceName+'-local')
	name = config.getstr('name', os.getenv('USER')) # os.getlogin() fails in minimal chroots
	email = config.getstr('email', os.getenv('USER')+'@'+os.uname()[1]) # user@hostname
	debDir = os.path.expanduser(config.getstr('debDir'))
	buildSystem = buildSystems[config.getstr('buildSystem')] # get the data immediately
	version = config.getstr('version')
	dbgPackage = config.getbool('dbgPackage', False)
	parallelJobs = config.getint('parallelJobs', 2)
	packageArchitecture = config.getstr('architecture', 'any')
	withPython2 = config.getbool('withPython2', False)
	if withPython2:
		buildSystem.binaryDepends.append("${python:Depends}") # HACK, but it works: make sure dependencies on binary are added
	# we return the list of files generated, so we need to know the architecture
	arch = getArchitecture(config)
	files = []
	# create folders
	if os.path.exists('debian'): raise Exception('debian folder already exists?')
	os.mkdir('debian')
	os.mkdir('debian/source')
	if not os.path.exists(debDir): os.makedirs(debDir)
	# source format file
	with open('debian/source/format', 'w') as f:
		print >>f, "3.0 (native)"
	# compat file
	with open('debian/compat', 'w') as f:
		print >>f, "9"
	# copyright file
	with open('debian/copyright', 'w') as f:
		print >>f, "Auto-generated by auto-debuild, not suited for distribution"
	# changelog file
	with open('debian/changelog', 'w') as f:
		print >>f, sourceName,"("+version+")","UNRELEASED; urgency=low"
		print >>f, ""
		print >>f, "  * Auto-generated by auto-debuild"
		print >>f, ""
		print >>f, " --",name,"<"+email+">  "+time.strftime('%a, %d %b %Y %H:%M:%S %z')
	# control file
	with open('debian/control', 'w') as f:
		print >>f, "Source:",sourceName
		print >>f, "Section:",config.getstr('section', 'misc')
		print >>f, "Priority: extra"
		print >>f, "Maintainer: %s <%s>" % (name, email)
		writeDependency(f, 'Build-Depends', ["debhelper (>= 9)"] + buildSystem.buildDepends + config.get('buildDepends', []))
		print >>f, "Standards-Version: 3.9.3"
		print >>f, ""
		print >>f, "Package:",binaryName
		print >>f, "Architecture:",packageArchitecture
		if 'binaryMultiArch' in config:
			print >>f, "Multi-Arch:",config.getstr('binaryMultiArch')
		writeDependency(f, "Pre-Depends", ["${misc:Pre-Depends}"] + config.get('binaryPreDepends', []))
		writeDependency(f, "Depends", ["${shlibs:Depends}", "${misc:Depends}"] + buildSystem.binaryDepends +
			config.get('binaryDepends', []))
		writeDependency(f, "Recommends", config.get('binaryRecommends', []))
		writeDependency(f, "Provides", config.get('binaryProvides', [sourceName]))
		writeDependency(f, "Conflicts", config.get('binaryConflicts', []))
		writeDependency(f, "Breaks", config.get('binaryBreaks', []) + config.get('binaryBreaksReplaces', []))
		writeDependency(f, "Replaces", config.get('binaryReplaces', []) + config.get('binaryBreaksReplaces', []))
		print >>f, "Description:",sourceName,"(auto-debuild)"
		print >>f, " Package auto-generated by auto-debuild."
		files.append(os.path.join(debDir, "%s_%s_%s.deb" % (binaryName, version, arch)))
		if dbgPackage:
			print >>f, ""
			print >>f, "Package:",binaryName+"-dbg"
			print >>f, "Section: debug"
			print >>f, "Priority: extra"
			print >>f, "Architecture:",packageArchitecture
			writeDependency(f, "Depends", ["${misc:Depends}", binaryName+" (= ${binary:Version})"])
			print >>f, "Description:",sourceName,"debug smbols (auto-debuild)"
			print >>f, " Package containing debug symbols for "+sourceName+", auto-generated by auto-debuild."
			files.append(os.path.join(debDir, "%s-dbg_%s_%s.deb" % (binaryName, version, arch)))
	# install file
	with open('debian/'+binaryName+'.install', 'w') as f:
		for line in config.get('binaryInstallFiles', []):
			if line.startswith('/'): # a file from within the package, not from the source tree
				line = 'debian/'+binaryName+line
			print >>f, line
	# maintainer scripts for alternatives
	if 'alternatives' in config:
		with open('debian/'+binaryName+'.postinst', 'w') as f:
			print >>f, "#!/bin/sh"
			print >>f, "set -e"
			print >>f, 'if [ "$1" = "configure" ]; then'
			for alternative in config.get('alternatives'):
				print >>f, safeCall('update-alternatives', '--install', alternative['link'], alternative['name'], alternative['target'],
					str(alternative['priority']))
			print >>f, 'fi'
			print >>f, ''
			print >>f, '#DEBHELPER#'
			print >>f, ''
			print >>f, 'exit 0'
		with open('debian/'+binaryName+'.prerm', 'w') as f:
			print >>f, "#!/bin/sh"
			print >>f, "set -e"
			print >>f, 'if [ "$1" = "remove" ]; then'
			for alternative in config.get('alternatives'):
				print >>f, safeCall('update-alternatives', '--remove', alternative['name'], alternative['target'])
			print >>f, 'fi'
			print >>f, ''
			print >>f, '#DEBHELPER#'
			print >>f, ''
			print >>f, 'exit 0'
	# rules file: build system specific
	with open('debian/rules', 'w') as f:
		# get rule file for build system: may only touch auto_config and auto_clean rules and the dh options
		r = buildSystem.ruleMaker(config)
		# global rules
		r.env["DEB_BUILD_OPTIONS"] = 'parallel='+str(parallelJobs)
		if not dbgPackage:
			# disable debug information
			r.env["DEB_CFLAGS_APPEND"] = '-g0'
			r.env["DEB_CXXFLAGS_APPEND"] = '-g0'
		r.dh += ['--parallel']
		if withPython2:
			r.dhWith.add('python2')
			r.rules['python2'] = ['dh_python2 --no-guessing-versions']
		r.rules['builddeb'] = [safeCall('dh_builddeb', "--destdir="+debDir)] # passing this gobally to dh results in weird problems (like stuff being installed there, and not in the package...)
		r.rules['auto_test'] = []
		# installation rule
		r.rules['auto_install'] = [safeCall('dh_auto_install', '--destdir=debian/'+binaryName)] # install everything into the binary package
		if 'binarySkipFiles' in config:
			r.rules['auto_install'].append(safeCall('cd', 'debian/'+binaryName) + " && " +
				safeCall('rm', *config.get('binarySkipFiles')))
		# for debug packages
		if dbgPackage:
			r.rules['strip'] = [safeCall('dh_strip', '--dbg-package='+binaryName+"-dbg")] # put debug files in appropriate package
			r.rules['installdocs'] = [safeCall('dh_installdocs', '--link-doc='+binaryName)] # make the doc folder of the dbg package a symlink
		# wait after configuration?
		if config.getbool('waitAfterConfig', False):
			if not 'auto_configure' in r.rules: r.rules['auto_configure'] = ['dh_auto_configure'] # make sure there is an override
			r.rules['auto_configure'].append("@"+safeCall('read', '-p', 'Configuration done. Hit "Enter" to build the package. ', 'DUMMY_VAR')) # if we run in dash, we need to tell it which variable to use for the result...
		# dump it to a file
		r.write(f)
	mode = os.stat('debian/rules').st_mode
	os.chmod('debian/rules', mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
	# return list of files that will be created
	return files

def buildDebianPackage(config):
	if not isinstance(config, ConfigDict):
		config = ConfigDict(config)
	commands = ['dpkg-checkbuilddeps', 'debian/rules clean', 'debian/rules build', 'fakeroot debian/rules binary', 'debian/rules clean']
	command = ['bash', '-c', ' && '.join(commands)] # make it all one command, so we don't have to open and close the chroot too often
	subprocess.check_call(commandInBuildEnv(config, command))
	shutil.rmtree('debian') # it only contains what we just created

###################################################################
# if we are called directly as script
if __name__ == "__main__":
	try:
		# get config
		config = loadConfigFile('auto-debuild.conf')
		# generate debian files
		if os.path.exists('debian'):
			if raw_input("A debian folder already exists, to you want to remove it (y/N)? ").lower() != "y":
				sys.exit(1)
			shutil.rmtree('debian')
		files = createDebianFiles(config)
		# check if a file is overwritten
		for file in files:
			if os.path.exists(file):
				if raw_input("Do you want to overwrite %s (y/N)? " % file).lower() != "y":
					sys.exit(1)
		# run compilation
		buildDebianPackage(config)
		# install files
		print "Installing created deb files..."
		subprocess.check_call(['sudo', 'dpkg', '--install'] + files)
	except (subprocess.CalledProcessError, KeyboardInterrupt) as e: # for some exceptions, a stackrace is usually pointless
		print >> sys.stderr
		print >> sys.stderr
		if isinstance(e, KeyboardInterrupt): # str(e) would be the empty string
			print >> sys.stderr, "Interruped by user"
		else:
			print >> sys.stderr, "Error during package creation: %s" % str(e)
		print >> sys.stderr
		sys.exit(1)
